#[cfg(feature = "axum")]
use axum::{
    extract::FromRequestParts,
    http::{StatusCode, request::Parts},
};
#[cfg(feature = "axum")]
use axum_extra::{
    TypedHeader,
    headers::{Authorization, authorization::Bearer},
};
use owl::db::Model;
#[cfg(feature = "rocket")]
use rocket::{Request, http::Status, outcome::Outcome};

use crate::{Sessions, User};

pub struct UserAuth(pub Model<User>);

#[cfg(feature = "axum")]
impl<S> FromRequestParts<S> for UserAuth
where
    S: Send + Sync,
{
    type Rejection = (StatusCode, &'static str);

    async fn from_request_parts(parts: &mut Parts, _state: &S) -> Result<Self, Self::Rejection> {
        let jar = axum_extra::extract::CookieJar::from_headers(&parts.headers);

        if let Some(cookie) = jar.get("session") {
            if let Some(user) = User::from_session(cookie.value().to_string()).await {
                return Ok(UserAuth(user));
            }
        }

        Err((StatusCode::UNAUTHORIZED, "Unauthorized"))
    }
}

#[cfg(feature = "rocket")]
#[rocket::async_trait]
impl<'r> rocket::request::FromRequest<'r> for UserAuth {
    type Error = ();

    async fn from_request(request: &'r Request<'_>) -> rocket::request::Outcome<Self, Self::Error> {
        if let Some(session_id) = request.cookies().get("session") {
            if let Some(user) = User::from_session(session_id.value().to_string()).await {
                return Outcome::Success(UserAuth(user));
            }
        }
        Outcome::Error((Status::Unauthorized, ()))
    }
}

/// Struct which extracts a user with session from `Token` HTTP Header.
pub struct APIUser(pub Model<User>);

#[cfg(feature = "rocket")]
#[rocket::async_trait]
impl<'r> rocket::request::FromRequest<'r> for APIUser {
    type Error = ();

    async fn from_request(request: &'r Request<'_>) -> rocket::request::Outcome<Self, Self::Error> {
        if let Some(auth_header) = request.headers().get_one("Authorization") {
            // Expect "Bearer <token>"
            if let Some(token) = auth_header.strip_prefix("Bearer ").map(str::trim) {
                if let Some(user) = User::from_session(token.to_string()).await {
                    return Outcome::Success(APIUser(user));
                }
            }
        }

        Outcome::Error((Status::Unauthorized, ()))
    }
}

#[cfg(feature = "axum")]
impl<S> FromRequestParts<S> for APIUser
where
    S: Send + Sync,
{
    type Rejection = (StatusCode, &'static str);

    async fn from_request_parts(parts: &mut Parts, state: &S) -> Result<Self, Self::Rejection> {
        let TypedHeader(Authorization(bearer)) =
            TypedHeader::<Authorization<Bearer>>::from_request_parts(parts, state)
                .await
                .map_err(|_| {
                    (
                        StatusCode::UNAUTHORIZED,
                        "Missing or invalid Authorization header",
                    )
                })?;

        let token = bearer.token();

        match User::from_session(token.to_string()).await {
            Some(user) => Ok(APIUser(user)),
            None => Err((StatusCode::UNAUTHORIZED, "Invalid token")),
        }
    }
}

/// Maybe User?
///
/// This struct extracts a user if possible, but also allows anybody.
///
/// # Example:
///
/// ```ignore
///
/// // Publicly accessable
/// #[get("/")]
/// pub async fn index(ctx: RequestContext, user: MaybeUser) -> StringResponse {
///     match user {
///         MaybeUser::User(user) => println!("You are {}", user.username),
///         MaybeUser::Anonymous => println!("Who are you?")
///     }
/// }
/// ```
pub enum MaybeUser {
    User(Model<User>),
    Anonymous,
}

#[cfg(feature = "rocket")]
#[rocket::async_trait]
impl<'r> rocket::request::FromRequest<'r> for MaybeUser {
    type Error = ();

    async fn from_request(request: &'r Request<'_>) -> rocket::request::Outcome<Self, Self::Error> {
        if let Some(session_id) = request.cookies().get("session") {
            if let Some(user) = User::from_session(session_id.value().to_string()).await {
                return Outcome::Success(MaybeUser::User(user));
            }
        }

        Outcome::Success(MaybeUser::Anonymous)
    }
}

#[cfg(feature = "axum")]
impl<S> FromRequestParts<S> for MaybeUser
where
    S: Send + Sync,
{
    type Rejection = (StatusCode, &'static str);

    async fn from_request_parts(parts: &mut Parts, _state: &S) -> Result<Self, Self::Rejection> {
        let jar = axum_extra::extract::CookieJar::from_headers(&parts.headers);

        if let Some(cookie) = jar.get("session") {
            if let Some(user) = User::from_session(cookie.value().to_string()).await {
                return Ok(MaybeUser::User(user));
            }
        }

        Ok(MaybeUser::Anonymous)
    }
}

impl From<MaybeUser> for Option<Model<User>> {
    fn from(value: MaybeUser) -> Self {
        value.take_user()
    }
}

impl MaybeUser {
    #[must_use]
    pub const fn user(&self) -> Option<&Model<User>> {
        match self {
            MaybeUser::User(user) => Some(user),
            MaybeUser::Anonymous => None,
        }
    }

    #[must_use]
    pub fn take_user(self) -> Option<Model<User>> {
        match self {
            MaybeUser::User(user) => Some(user),
            MaybeUser::Anonymous => None,
        }
    }
}

/// Admin User
///
/// This struct expects an Admin User and returns `Forbidden` otherwise.
///
/// # Example:
///
/// ```ignore
///
/// // Only admin users can access this route
/// #[get("/admin")]
/// pub async fn admin_panel(ctx: RequestContext, user: AdminUser) -> StringResponse {
///     ...
/// }
/// ```
pub struct AdminUser(pub Model<User>);

#[cfg(feature = "rocket")]
#[rocket::async_trait]
impl<'r> rocket::request::FromRequest<'r> for AdminUser {
    type Error = ();

    async fn from_request(request: &'r Request<'_>) -> rocket::request::Outcome<Self, Self::Error> {
        if let Some(session_id) = request.cookies().get("session") {
            if let Some(user) = User::from_session(session_id.value().to_string()).await {
                if user.read().is_admin() {
                    return Outcome::Success(AdminUser(user));
                }
            }
        }

        Outcome::Error((Status::Unauthorized, ()))
    }
}

#[cfg(feature = "axum")]
impl<S> FromRequestParts<S> for AdminUser
where
    S: Send + Sync,
{
    type Rejection = (StatusCode, &'static str);

    async fn from_request_parts(parts: &mut Parts, _state: &S) -> Result<Self, Self::Rejection> {
        let jar = axum_extra::extract::CookieJar::from_headers(&parts.headers);

        if let Some(cookie) = jar.get("session") {
            if let Some(user) = User::from_session(cookie.value().to_string()).await {
                if user.read().is_admin() {
                    return Ok(AdminUser(user));
                }
            }
        }

        Err((StatusCode::UNAUTHORIZED, "Unauthorized"))
    }
}
