use owl::{db::model::file::File, get, prelude::*, query, save, update};
use serde::{Deserialize, Serialize};

// TODO : 2FA

/// User
#[derive(Debug)]
#[model]
pub struct User {
    /// The username chosen by the user
    pub id: Id,
    /// The hashed password for the user
    pub password: String,
    /// The role of the user
    pub user_role: UserRole,
    #[serde(skip)]
    pub session: String,
    pub profile_picture: Option<IdRef<File>>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum UserRole {
    /// A regular user with limited permissions
    Regular,
    /// An admin user with full system privileges
    Admin,
}

impl User {
    /// Find a user by their username
    pub async fn find(username: &str) -> Option<Model<Self>> {
        get!(username)
    }

    /// Create a new user with the given details
    ///
    /// Returns an Option containing the created user, or None if a user already exists with the same username
    pub async fn create(username: String, password: &str, role: UserRole) -> Option<Model<Self>> {
        // Check if a user already exists with the same username
        if Self::find(&username).await.is_some() {
            return None;
        }

        let u = Self {
            id: Id::String(username),
            password: bcrypt::hash(password, bcrypt::DEFAULT_COST).unwrap(),
            user_role: role,
            profile_picture: None,
            session: String::new(),
        };

        Some(save!(u))
    }

    /// Change the password of a User
    ///
    /// Returns a Result indicating whether the password change was successful or not
    pub async fn passwd(&self, old: &str, new: &str) -> Result<(), ()> {
        if self.verify_pw(old) {
            let mut target = vec![get!(self.id.clone()).unwrap()];
            update!(&mut target, |u: &mut User| {
                u.password = bcrypt::hash(new, bcrypt::DEFAULT_COST).unwrap();
            });

            return Ok(());
        }

        Err(())
    }

    /// Find all users in the system
    #[must_use]
    pub async fn find_all() -> Vec<Model<Self>> {
        query!(|_| true)
    }

    /// Check if the user is an admin
    #[must_use]
    pub const fn is_admin(&self) -> bool {
        matches!(self.user_role, UserRole::Admin)
    }

    /// Verify that a provided password matches the hashed password for the user
    ///
    /// Returns a boolean indicating whether the passwords match or not
    #[must_use]
    pub fn verify_pw(&self, password: &str) -> bool {
        bcrypt::verify(password, &self.password).unwrap()
    }
}
